package at.gv.landooe.shop.dao;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ConcurrentNavigableMap;

import org.apache.log4j.Logger;
import org.mapdb.DB;

import at.gv.landooe.shop.elems.Item;
import at.gv.landooe.shop.elems.Order;
import at.gv.landooe.shop.elems.OrderItem;

/**
 * DAO Stub that provides an interface similar to DAO patterns in
 * well known frameworks (Spring, etc.) however is only a very thin
 * layer on top of a MapDB database stored in the users $TMP directory.  
 * 
 * @author matthias
 */
public class ShopDAO {
	
	public static final String ITEMS = "items";
	public static final String ORDERS = "orders";
	
	static Logger logger = Logger.getLogger(ShopDAO.class);
	
	// ###############
	// ITEMS
	// ###############

	/**
	 * Find an item with a specific ID in the database
	 * 
	 * @param id
	 * @return
	 */
	public static Item getItem(int id) {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Item> map = db.getTreeMap(ITEMS);
		return map.get(id);
	}
	
	/**
	 * Save the item
	 * 
	 * @param item
	 * @return
	 */
	public static Item saveItem(Item item) {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Item> map = db.getTreeMap(ITEMS);
		if(item.getId() <= 0) {
			Integer lastKey = map.lastKey();
			item.setId(lastKey + 1);
		}
		map.put(item.getId(), item);
		db.commit();
		return item;
	}
	
	/**
	 * Delete item; ignore deletes  on non-existent item
	 * 
	 * @param item
	 */
	public static void deleteItem(Item item) {
		deleteItem(item.getId());
	}
	
	/**
	 * Delete item by id
	 * 
	 * @param id
	 */
	public static void deleteItem(int id) {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Item> map = db.getTreeMap(ITEMS);
		map.remove(id);
		db.commit();
	}
	
	/**
	 * Retrieve all {@link Item}s from the database
	 * 
	 * @return
	 */
	public static List<Item> listItems() {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Item> map = db.getTreeMap(ITEMS);
		ArrayList<Item> items = new ArrayList<Item>();
		items.addAll(map.values());
		return items;
	}
	
	// ###############
	// ITEMS
	// ###############
	
	/**
	 * Retrieves the list of all currently stored orders
	 * 
	 * @return
	 */
	public static List<Order> listOrders() {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Order> map = db.getTreeMap(ORDERS);
		ArrayList<Order> orders = new ArrayList<>();
		orders.addAll(map.values());
		return orders;
	}
	
	/**
	 * Saves an order
	 * 
	 * @param order
	 * @return
	 */
	public static Order saveOrder(Order order) {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Order> map = db.getTreeMap(ORDERS);
		Order storedOrder = null;
		if(order.getId() <= 0) {
			Integer lastKey = map.lastKey();
			order.setId(lastKey +1);
		}else{
			storedOrder = map.get(order.getId());
		}
		updateStockQuantities(db, order, storedOrder);
		map.put(order.getId(), order);
		db.commit();
		return order;
	}
	
	/**
	 * Retrieve a single order
	 * 
	 * @param id
	 * @return
	 */
	public static Order getOrder(int id) {
		DB db = MapDBDatabaseConnection.getDb();
		ConcurrentNavigableMap<Integer, Order> map = db.getTreeMap(ORDERS);
		return map.get(id);
	}
	
	/**
	 * 
	 * 
	 * @param db
	 * @param order
	 */
	private static void updateStockQuantities(DB db, Order order, Order storedOrder) {
		ConcurrentNavigableMap<Integer, Item> map = db.getTreeMap(ITEMS);
		for(OrderItem oi : order.getItems()) {
			Item item = map.get(oi.getItem());
			if(item != null) {
				OrderItem storedItem = null;
				if(storedOrder != null) {
					storedItem = storedOrder.findOrderItem(oi.getItem());
				}
				int diffQuantity;
				if(storedItem != null) {
					diffQuantity = oi.getQuantity() - storedItem.getQuantity();
				}else {
					diffQuantity = oi.getQuantity();
				}
				item.setStock(item.getStock() - diffQuantity);
				map.put(item.getId(), item);
			}
		}
		db.commit();
	}
	
	/**
	 * Populates the database DB with test data
	 * 
	 * method is package protected on purpose > allows calling from {@link MapDBDatabaseConnection}
	 * 
	 * @param db
	 */
	static void runPopulateScenario(DB db) {
		logger.info("Populating database with testdata");
		
		ConcurrentNavigableMap<Integer, Item> itemMap = db.getTreeMap(ITEMS);
		
		itemMap.put(1, new Item(1, "Raspberry Pi Model 2", 55f, 124));
		itemMap.put(2, new Item(2, "Raspberry Pi Model A", 25f, 4));
		itemMap.put(3, new Item(3, "Raspberry Pi Model B", 35f, 7));
		itemMap.put(4, new Item(4, "MicroSD Card 32 GB", 27f, 153));
		
		logger.info("Successfully added " + itemMap.size() + " items");
		
		ConcurrentNavigableMap<Integer, Order> orderMap = db.getTreeMap(ORDERS);
		
		Order o;
		
		o = new Order(1, "Matthias Steinbauer");
		o.getItems().add(new OrderItem(1, 10, 53f));
		o.getItems().add(new OrderItem(4, 10, 25f));
		orderMap.put(o.getId(), o);
		
		o = new Order(2, "Thomas Tester");
		o.getItems().add(new OrderItem(3, 1, 35f));
		o.getItems().add(new OrderItem(4, 1, 27f));
		orderMap.put(o.getId(), o);
		
		o = new Order(3, "Julian JUnit-Tester");
		o.getItems().add(new OrderItem(2, 2, 25f));
		orderMap.put(o.getId(), o);
		
		logger.info("Successfully added " + orderMap.size() + " orders");
		
		db.commit();
		
	}

}
